"""Mirakuru exceptions."""


class ExecutorError(Exception):
    """Base exception for executor failures."""

    def __init__(self, executor):
        """
        Exception initialization.

        :param mirakuru.base.Executor executor: for which exception occurred
        """
        super(ExecutorError, self).__init__(self)
        self.executor = executor


class TimeoutExpired(ExecutorError):
    """Is raised when the timeout expires while starting an executor."""

    def __init__(self, executor, timeout):
        """
        Exception initialization with an extra ``timeout`` argument.

        :param mirakuru.base.Executor executor: for which exception occurred
        :param int timeout: timeout for which exception occurred
        """
        super(TimeoutExpired, self).__init__(executor)
        self.timeout = timeout

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return 'Executor {0} timed out after {1} seconds'.format(
            self.executor, self.timeout
        )


class AlreadyRunning(ExecutorError):
    """
    Is raised when the executor seems to be already running.

    When some other process (not necessary executor) seems to be started with
    same configuration we can't bind to same port.
    """

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return ("Executor {exc.executor} seems to be already running. "
                "It looks like the previous executor process hasn't been "
                "terminated or killed. Also there might be some completely "
                "different service listening on {exc.executor.port} port."
                .format(exc=self))


class ProcessExitedWithError(ExecutorError):
    """
    Raised when the process invoked by the executor returns a non-zero code.

    We allow the process to exit with zero because we support daemonizing
    subprocesses. We assume that when double-forking, the parent process will
    exit with 0 in case of successful daemonization.
    """

    def __init__(self, executor, exit_code):
        """
        Exception initialization with an extra ``exit_code`` argument.

        :param mirakuru.base.Executor executor: for which exception occurred
        :param int exit_code: code the subprocess exited with
        """
        super(ProcessExitedWithError, self).__init__(executor)
        self.exit_code = exit_code

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return ("The process invoked by the {exc.executor} executor has "
                "exited with a non-zero code: {exc.exit_code}."
                .format(exc=self))



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""HTTP enabled process executor."""

import re
import socket

from mirakuru.compat import HTTPConnection, HTTPException
from mirakuru.compat import urlparse

from mirakuru.tcp import TCPExecutor


class HTTPExecutor(TCPExecutor):
    """Http enabled process executor."""

    DEFAULT_PORT = 80
    """Default TCP port for the HTTP protocol."""

    def __init__(self, command, url, status=r'^2\d\d$', **kwargs):
        """
        Initialize HTTPExecutor executor.

        :param (str, list) command: command to be run by the subprocess
        :param str url: URL that executor checks to verify
            if process has already started.
        :param bool shell: same as the `subprocess.Popen` shell definition
        :param str|int status: HTTP status code(s) that an endpoint must
            return for the executor being considered as running. This argument
            is interpreted as a single status code - e.g. '200' or '404' but
            also it can be a regular expression - e.g. '4..' or '(200|404)'.
            Default: any 2XX HTTP status code.
        :param int timeout: number of seconds to wait for the process to start
            or stop. If None or False, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        :param int sig_stop: signal used to stop process run by the executor.
            default is `signal.SIGTERM`
        :param int sig_kill: signal used to kill process run by the executor.
            default is `signal.SIGKILL`

        """
        self.url = urlparse(url)
        """
        An :func:`urlparse.urlparse` representation of an url.

        It'll be used to check process status on.
        """

        port = self.url.port
        if port is None:
            port = self.DEFAULT_PORT

        self.status = str(status)
        self.status_re = re.compile(str(status))

        super(HTTPExecutor, self).__init__(
            command, host=self.url.hostname, port=port, **kwargs
        )

    def after_start_check(self):
        """Check if defined URL returns expected status to a HEAD request."""
        try:
            conn = HTTPConnection(self.host, self.port)

            conn.request('HEAD', self.url.path)
            status = str(conn.getresponse().status)

            if status == self.status or self.status_re.match(status):
                conn.close()
                return True

        except (HTTPException, socket.timeout, socket.error):
            return False
